/* Exercice 3. */

/* The graph. */
arcs([[a,b,5], [a,c,3], [b,d,2], [c,d,5], [d,e,2]]).

/* The solution works a lot with lists containing elements [Key,Value],
   called "hash" in the following. */

/* Find the value for a given key in a hash. */
value([[K,V]|_],K,V).
value([[K1,_]|H],K,V) :- K1\=K, value(H,K,V).

/* update(H1,K,V,H2) - update hash H1 with a new value V for key K,
	return result in H2; if K does not exist in H1, it will be added. */
update([],K,V,[[K,V]]).
update([[K,_]|H],K,V,[[K,V]|H]).
update([[K1,V1]|H1],K,V,[[K1,V1]|H2]) :- K1\=K, update(H1,K,V,H2).

/* filter(H1,List,H2) removes all pairs from hash H1 such that their key
	is in List, putting the result into H2 */
filter([],_,[]).
filter([[K,V]|H1],List,[[K,V]|H2])
	:- \+member(K,List),filter(H1,List,H2).
filter([[K,_]|H1],List,H2)
	:- member(K,List),filter(H1,List,H2).

/* minValue(H,K,V) finds the key-value pair with minimal value in H */
minValue([[K,V]],K,V).
minValue([[_,V1]|H],K,V) :- minValue(H,K,V), V =< V1.
minValue([[K1,V1]|H],K1,V1) :- minValue(H,_,V), V > V1.

/* like minValue, but K must not be in List */
minValueFilter(H,List,K,V) :- filter(H,List,H1), minValue(H1,K,V).

/* changeDist(S,A,D1,P1,D2,P2) takes the following arguments:
	node S;
	arc list A;
	D1 are hashes giving the current distance for each node;
	P1 are hashes giving the current path from the source to each node.
   Then changeDist takes all arcs starting at S, and if using these arcs
   provides shorter paths than those that are known, updates D1,P1 to D2,P2. */

/* no arcs left, nothing changes */
changeDist(_,[],Dist,PathTo,Dist,PathTo).

/* an arc starting with a node different from S, skip it */
changeDist(S,[[T,_,_]|A],Dist,PathTo,Dist1,PathTo1)
	:- S \= T, changeDist(S,A,Dist,PathTo,Dist1,PathTo1).

/* an arc S->T giving a path that is longer than the known path to T */
changeDist(S,[[S,T,D]|A],Dist,PathTo,Dist1,PathTo1)
	:- value(Dist,S,D1),value(Dist,T,D2), D1+D >= D2,
		changeDist(S,A,Dist,PathTo,Dist1,PathTo1).

/* an arc to a node which was not reachable before */
changeDist(S,[[S,T,D]|A],Dist,PathTo,Dist1,PathTo1)
	:- value(Dist,S,D1),\+value(Dist,T,_), DD is D1+D,
		update(Dist,T,DD,Dist2),
		value(PathTo,S,P),append(P,[T],P1),update(PathTo,T,P1,PathTo2),
		changeDist(S,A,Dist2,PathTo2,Dist1,PathTo1).

/* an arc giving a shorter path than the previous best path to T */
changeDist(S,[[S,T,D]|A],Dist,PathTo,Dist1,PathTo1)
	:- value(Dist,S,D1),value(Dist,T,D2), DD is D1+D, DD < D2, 
		update(Dist,T,DD,Dist2),
		value(PathTo,S,P),append(P,[T],P1),update(PathTo,T,P1,PathTo2),
		changeDist(S,A,Dist2,PathTo2,Dist1,PathTo1).

/* The main procedure way(S,T,D,P,Done): Find path from source to target,
   where D,P are hashes giving distances and shortest path to the nodes
   in Done, the latter being the list of nodes that have been processed
   and whose shortest path is thus already known. */

/* if we've found the target node, output the result */
way(S,T,Dist,PathTo,_)
	:- S=T, value(PathTo,T,P), value(Dist,T,D), write(D), nl, write(P), nl.
/* otherwise process outgoing arcs from S, update the hashes, pick the remaining
   node with shortest distance and iterate. */
way(S,T,Dist,PathTo,Done)
	:- S\=T, arcs(A), changeDist(S,A,Dist,PathTo,Dist1,PathTo1),
		minValueFilter(Dist1,Done,U,_), way(U,T,Dist1,PathTo1,[U|Done]).

/* invocation predicate: source S, target T */
solve(S,T) :- way(S,T,[[S,0]],[[S,[S]]],[S]).
